"""
Host cli module
"""
import click

from pyepp.cli import utils
from pyepp.host import Host, IPAddressData, HostData


@click.group(name='host')
@click.pass_context
def host_group(ctx):
    """To work with Host objects in the registry."""
    ctx.obj.registry_object = Host(ctx.obj.epp)


@click.command(name='info')
@click.argument('host-name')
@click.option('--client-transaction-id')
@click.pass_context
def host_info(ctx, host_name, client_transaction_id) -> None:
    """Returns host details.

    HOST_NAME: Host name
    """
    result = ctx.obj.info(host_name, client_transaction_id)
    utils.echo(result)


@click.command(name='check')
@click.argument('host-names', nargs=-1)
@click.option('--client-transaction-id')
@click.pass_context
def host_check(ctx, host_names: tuple, client_transaction_id) -> None:
    """Checks if host(s) exist in the registry."""
    result = ctx.obj.check(list(host_names), client_transaction_id)
    utils.echo(result)


@click.command(name='delete')
@click.argument('host-name')
@click.option('--client-transaction-id')
@click.pass_context
def host_delete(ctx, host_name, client_transaction_id) -> None:
    """Deletes a given host from registry.

    HOST_NAME: Host name
    """
    result = ctx.obj.delete(host_name, client_transaction_id)
    utils.echo(result)


@click.command(name='update')
@click.argument('host-name')
@click.option('--add-ip', multiple=True, nargs=2, type=click.Tuple([str, str]),
              help="Add IP address to host. <ADDRESS IP[v4, v6]>")
@click.option('--remove-ip', multiple=True, nargs=2, type=click.Tuple([str, str]),
              help="Remove IP address to host. <ADDRESS IP[v4, v6]>")
@click.option('--add-status', multiple=True, help="Add a status to host.")
@click.option('--remove-status', multiple=True, help="Remove a status to host.")
@click.option('--new-host-name', help="The host name will be changed to this new host name.")
@click.option('--client-transaction-id')
@click.pass_context
# pylint: disable=too-many-arguments, too-many-locals, too-many-boolean-expressions
def host_update(ctx, host_name, add_ip, remove_ip, add_status, remove_status, new_host_name,
                client_transaction_id) -> None:
    """Updates host's details.

    HOST_NAME: Host name
    """
    add_ip_address = [IPAddressData(item[0], item[1]) for item in add_ip] if add_ip else None
    remove_ip_address = [IPAddressData(item[0], item[1]) for item in remove_ip] if add_ip else None
    add_statue = list(add_status) if add_status else None
    remove_status = list(remove_status) if add_status else None
    result = ctx.obj.update(host_name,
                            add_ip_address=add_ip_address,
                            remove_ip_address=remove_ip_address,
                            add_status=add_statue,
                            remove_status=remove_status,
                            new_host_name=new_host_name,
                            client_transaction_id=client_transaction_id)
    utils.echo(result)


@click.command(name='create')
@click.argument('host-name')
@click.option('--ip-address', multiple=True, nargs=2, type=click.Tuple([str, str]),
              help="Add IP address to host. At least one required. <ADDRESS IP[v4, v6]>", required=True)
@click.option('--client-transaction-id')
@click.pass_context
def host_create(ctx, host_name, ip_address, client_transaction_id) -> None:
    """Creates a new contact in the registry.

    HOST_NAME: A unique host name
    """
    ip_address = [IPAddressData(item[0], item[1]) for item in ip_address] if ip_address else None
    result = ctx.obj.create(HostData(host_name=host_name,
                                     address=ip_address),
                            client_transaction_id)
    utils.echo(result)


host_group.add_command(host_info)
host_group.add_command(host_check)
host_group.add_command(host_delete)
host_group.add_command(host_update)
host_group.add_command(host_create)



"""
EPP XML command templates
"""
from jinja2 import Environment, BaseLoader

template_engine = Environment(loader=BaseLoader(),
                              trim_blocks=True,
                              lstrip_blocks=True,
                              autoescape=True)

HELLO_XML = """<?xml version="1.0" encoding="UTF-8"?>
<epp xmlns="urn:ietf:params:xml:ns:epp-1.0">
  <hello/>
</epp>"""

LOGIN_XML = """<?xml version="1.0" encoding="UTF-8" standalone="no"?>
<epp xmlns="urn:ietf:params:xml:ns:epp-1.0">
  <command>
    <login>
      <clID>{{ user }}</clID>
      <pw>{{ password }}</pw>
      <options>
        <version>1.0</version>
        <lang>en</lang>
      </options>
      <svcs>
        <objURI>urn:ietf:params:xml:ns:domain-1.0</objURI>
        <objURI>urn:ietf:params:xml:ns:contact-1.0</objURI>
        <objURI>urn:ietf:params:xml:ns:host-1.0</objURI>
        <svcExtension>
          <extURI>urn:ietf:params:xml:ns:secDNS-1.1</extURI>
          {% for extension in extensions %}
          <extURI>{{ extension }}</extURI>
          {% endfor %}
        </svcExtension>
      </svcs>
    </login>
  </command>
</epp>"""

LOGOUT_XML = """<?xml version="1.0" encoding="UTF-8"?>
<epp xmlns="urn:ietf:params:xml:ns:epp-1.0" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
xsi:schemaLocation="urn:ietf:params:xml:ns:epp-1.0 epp-1.0.xsd">
    <command>
        <logout/>
    </command>
</epp>"""

CONTACT_CHECK_XML = """<?xml version="1.0" encoding="UTF-8" standalone="no"?>
<epp xmlns="urn:ietf:params:xml:ns:epp-1.0">
  <command>
    <check>
      <contact:check xmlns:contact="urn:ietf:params:xml:ns:contact-1.0">
      {% for id in ids %}
        <contact:id>{{ id }}</contact:id>
      {% endfor %}
      </contact:check>
    </check>
    <clTRID>{{ client_transaction_id }}</clTRID>
  </command>
</epp>"""

CONTACT_INFO_XML = """<?xml version="1.0" encoding="UTF-8" standalone="no"?>
<epp xmlns="urn:ietf:params:xml:ns:epp-1.0">
  <command>
    <info>
      <contact:info xmlns:contact="urn:ietf:params:xml:ns:contact-1.0">
        <contact:id>{{ id }}</contact:id>
      </contact:info>
    </info>
    <clTRID>{{ client_transaction_id }}</clTRID>
  </command>
</epp>
"""

CONTACT_CREAT_XML = """<?xml version="1.0" encoding="UTF-8" standalone="no"?>
<epp xmlns="urn:ietf:params:xml:ns:epp-1.0">
  <command>
    <create>
      <contact:create xmlns:contact="urn:ietf:params:xml:ns:contact-1.0">
        <contact:id>{{ id }}</contact:id>
          <contact:postalInfo type="loc">
            <contact:name>{{ name }}</contact:name>
            {% if organization %} <contact:org>{{ organization }}</contact:org> {% endif %}
            <contact:addr>
              <contact:street>{{ street_1 }}</contact:street>
              {% if street_2 %} <contact:street>{{ street_2 }}</contact:street> {% endif %}
              {% if street_3 %} <contact:street>{{ street_3 }}</contact:street> {% endif %}
              <contact:city>{{ city }}</contact:city>
              {% if province %} <contact:sp>{{ province }}</contact:sp> {% endif %}
              {% if postal_code %} <contact:pc>{{ postal_code }}</contact:pc> {% endif %}
              <contact:cc>{{ country_code }}</contact:cc>
           </contact:addr>
         </contact:postalInfo>
         {% if phone %} <contact:voice>{{ phone }}</contact:voice> {% endif %}
         {% if fax %} <contact:fax>{{ fax }}</contact:fax> {% endif %}
         <contact:email>{{ email }}</contact:email>
         <contact:authInfo>
           <contact:pw> {{ password }}</contact:pw>
         </contact:authInfo>
         {% if privacy %}
         <contact:disclose flag="0">
           {% for element in privacy %}
           <contact:{{ element }}/>
           {% endfor %}
         </contact:disclose>
         {% endif %}
       </contact:create>
    </create>
    <clTRID>{{ client_transaction_id }}</clTRID>
  </command>
</epp>"""

CONTACT_DELETE_XML = """<?xml version="1.0" encoding="UTF-8" standalone="no"?>
<epp xmlns="urn:ietf:params:xml:ns:epp-1.0">
  <command>
    <delete>
      <contact:delete xmlns:contact="urn:ietf:params:xml:ns:contact-1.0">
        <contact:id>{{ id }}</contact:id>
      </contact:delete>
    </delete>
    <clTRID>{{ client_transaction_id }}</clTRID>
  </command>
</epp>"""

CONTACT_UPDATE_XML = """<?xml version="1.0" encoding="UTF-8" standalone="no"?>
<epp xmlns="urn:ietf:params:xml:ns:epp-1.0">
  <command>
    <update>
      <contact:update xmlns:contact="urn:ietf:params:xml:ns:contact-1.0">
        <contact:id>{{ id }}</contact:id>
        {% if add_status %}
        <contact:add>
            <contact:status s={{ add_status }}/>
        </contact:add>
        {% endif %}
        {% if remove_status %}
        <contact:rem>
            <contact:status s={{ remove_status }}/>
        </contact:rem>
        {% endif %}
        <contact:chg>
         {% if postalinfo_change %}
          <contact:postalInfo type="loc">
            {% if name %} <contact:name>{{ name }}</contact:name> {% endif %}
            {% if organization %} <contact:org>{{ organization }}</contact:org> {% endif %}
            {% if address_change %}
            <contact:addr>
                {% if street_1 %} <contact:street>{{ street_1 }}</contact:street> {% endif %}
                {% if street_2 %} <contact:street>{{ street_2 }}</contact:street> {% endif %}
                {% if street_3 %} <contact:street>{{ street_3 }}</contact:street> {% endif %}
                <contact:city>{{ city }}</contact:city>
                {% if province %} <contact:sp>{{ province }}</contact:sp> {% endif %}
                {% if postal_code %} <contact:pc>{{ postal_code }}</contact:pc> {% endif %}
                <contact:cc>{{ country_code }}</contact:cc>
           </contact:addr>
           {% endif %}
         </contact:postalInfo>
         {% endif %}
         {% if phone %} <contact:voice>{{ phone }}</contact:voice> {% endif %}
         {% if fax %} <contact:fax>{{ fax }}</contact:fax> {% endif %}
         {% if email %} <contact:email>{{ email }}</contact:email> {% endif %}
         <contact:authInfo>
           <contact:pw>{{ password }}</contact:pw>
         </contact:authInfo>
         {% if privacy %}
         <contact:disclose flag="0">
           {% for element in privacy %}
           <contact:{{ element }}/>
           {% endfor %}
         </contact:disclose>
         {% endif %}
        </contact:chg>
       </contact:update>
    </update>
    <clTRID>{{ client_transaction_id }}</clTRID>
  </command>
</epp>"""

DOMAIN_CHECK_XML = """<?xml version="1.0" encoding="UTF-8" standalone="no"?>
<epp xmlns="urn:ietf:params:xml:ns:epp-1.0">
  <command>
    <check>
      <domain:check xmlns:domain="urn:ietf:params:xml:ns:domain-1.0">
      {% for domain_name in domain_names %}
        <domain:name>{{ domain_name }}</domain:name>
      {% endfor %}
      </domain:check>
    </check>
    <clTRID>{{ client_transaction_id }}</clTRID>
  </command>
</epp>"""

DOMAIN_INFO_XML = """<?xml version="1.0" encoding="UTF-8" standalone="no"?>
<epp xmlns="urn:ietf:params:xml:ns:epp-1.0">
  <command>
    <info>
      <domain:info xmlns:domain="urn:ietf:params:xml:ns:domain-1.0">
        <domain:name hosts="all">{{ domain_name }}</domain:name>
        {% if udai %}
        <domain:authInfo>
          <domain:pw>{{ udai }}</domain:pw>
        </domain:authInfo>
        {% endif %}
      </domain:info>
    </info>
    <clTRID>{{ client_transaction_id }}</clTRID>
  </command>
</epp>"""

DOMAIN_CREATE_XML = """<?xml version="1.0" encoding="UTF-8" standalone="no"?>
<epp xmlns="urn:ietf:params:xml:ns:epp-1.0">
  <command>
    <create>
      <domain:create xmlns:domain="urn:ietf:params:xml:ns:domain-1.0">
        <domain:name>{{ domain_name }}</domain:name>
        <domain:period unit='y'>{{ period }}</domain:period>
        {% if host %}
        <domain:ns>
        {% for ns in host %}
          <domain:hostObj>{{ ns }}</domain:hostObj>
        {% endfor %}
        </domain:ns>
        {% endif %}
        <domain:registrant>{{ registrant }}</domain:registrant>
        {% if admin %}
        <domain:contact type="admin">{{ admin }}</domain:contact>
        {% endif %}
        {% if tech %}
        <domain:contact type="tech">{{ tech }}</domain:contact>
        {% endif %}
        {% if billing %}
        <domain:contact type="billing">{{ billing }}</domain:contact>
        {% endif %}
        <domain:authInfo>
          <domain:pw>{{ password }}</domain:pw>
        </domain:authInfo>
      </domain:create>
    </create>
    {% if dns_sec %}
    <extension>
      <secDNS:create xmlns:secDNS="urn:ietf:params:xml:ns:secDNS-1.1">
        <secDNS:dsData>
          <secDNS:keyTag>{{ dns_sec.get('key_tag') }}</secDNS:keyTag>
          <secDNS:alg>{{ dns_sec.get('algorithm') }}</secDNS:alg>
          <secDNS:digestType>{{ dns_sec.get('digest_type') }}</secDNS:digestType>
          <secDNS:digest>{{ dns_sec.get('digest') }}</secDNS:digest>
          {% if dns_sec.get('dns_key') %}
          <secDNS:keyData>
            <secDNS:flags>{{ dns_sec.get('dns_key').get('flag') }}</secDNS:flags>
            <secDNS:protocol>{{ dns_sec.get('dns_key').get('protocol') }}</secDNS:protocol>
            <secDNS:alg>{{ dns_sec.get('dns_key').get('algorithm') }}</secDNS:alg>
            <secDNS:pubKey>{{ dns_sec.get('dns_key').get('public_key') }}</secDNS:pubKey>
          </secDNS:keyData>
          {% endif %}
        </secDNS:dsData>
        {% if dns_key %}
        <secDNS:keyData>
          <secDNS:flags>{{ dns_key.dns_key.get('flag') }}</secDNS:flags>
          <secDNS:protocol>{{ dns_key.dns_key.get('protocol') }}</secDNS:protocol>
          <secDNS:alg>{{ dns_key.dns_key.get('algorithm') }}</secDNS:alg>
          <secDNS:pubKey>{{ dns_key.dns_key.get('public_key') }}</secDNS:pubKey>
        </secDNS:keyData>
        {% endif %}
      </secDNS:create>
    </extension>
    {% endif %}
    <clTRID>{{ client_transaction_id }}</clTRID>
  </command>
</epp>"""

DOMAIN_DELETE_XML = """<?xml version="1.0" encoding="UTF-8"?>
<epp xmlns="urn:ietf:params:xml:ns:epp-1.0">
  <command>
    <delete>
      <domain:delete xmlns:domain="urn:ietf:params:xml:ns:domain-1.0">
        <domain:name>{{ domain_name }}</domain:name>
      </domain:delete>
    </delete>
    <clTRID>{{ client_transaction_id }}</clTRID>
  </command>
</epp>"""

DOMAIN_RENEW_XML = """<?xml version="1.0" encoding="UTF-8"?>
<epp xmlns="urn:ietf:params:xml:ns:epp-1.0">
  <command>
    <renew>
      <domain:renew xmlns:domain="urn:ietf:params:xml:ns:domain-1.0">
        <domain:name>{{ domain_name }}</domain:name>
        <domain:curExpDate>{{ expiry_date }}</domain:curExpDate>
        {% if period %}
        <domain:period unit='y'>{{ period }}</domain:period>
        {% endif %}
      </domain:renew>
    </renew>
    <clTRID>{{ client_transaction_id }}</clTRID>
  </command>
</epp>"""

TRANSFER_REQUEST_XML = """<?xml version="1.0" encoding="UTF-8"?>
<epp xmlns="urn:ietf:params:xml:ns:epp-1.0">
<command>
  <transfer op="request">
    <domain:transfer xmlns:domain="urn:ietf:params:xml:ns:domain-1.0">
      <domain:name>{{ domain_name }}</domain:name>
      {% if period %}
      <domain:period unit='y'>{{ period }}</domain:period>
      {% endif %}
      <domain:authInfo>
        <domain:pw>{{ password }}</domains:pw>
      </domain:authInfo>
    </domain:transfer>
  </transfer>
  <clTRID>{{ client_transaction_id }}</clTRID>
</command>
</epp>"""

DOMAIN_UPDATE_XML = """<?xml version="1.0" encoding="UTF-8"?>
<epp xmlns="urn:ietf:params:xml:ns:epp-1.0">
  <command>
    <update>
      <domain:update xmlns:domain="urn:ietf:params:xml:ns:domain-1.0">
        <domain:name>{{ domain_name }}</domain:name>
        {% if add %}
        <domain:add>
          {% for admin in add_admins %}
          <domain:contact type="admin">{{ admin }}</domain:contact>
          {% endfor %}
          {% for tech in add_techs %}
          <domain:contact type="tech">{{ tech }}</domain:contact>
          {% endfor %}
          {% for billing in add_billings %}
          <domain:contact type="billing">{{ billing }}</domain:contact>
          {% endfor %}
          {% for status in add_statues %}
          <domain:status s="{{ status[0] }}" lang="en">{{ status[1] }}</domain:status>
          {% endfor %}
          {% if add_hosts %}
          <domain:ns>
            {% for host in add_hosts %}
            <domain:hostObj>{{ host }}</domain:hostObj>
            {% endfor %}
          </domain:ns>
          {% endif %}
        </domain:add>
        {% endif %}
        {% if remove %}
        <domain:rem>
          {% for admin in remove_admins %}
          <domain:contact type="admin">{{ admin }}</domain:contact>
          {% endfor %}
          {% for tech in remove_techs %}
          <domain:contact type="tech">{{ tech }}</domain:contact>
          {% endfor %}
          {% for billing in remove_billings %}
          <domain:contact type="billing">{{ billing }}</domain:contact>
          {% endfor %}
          {% for status in remove_statues %}
          <domain:status s="{{ status }}" />
          {% endfor %}
          {% if remove_hosts %}
          <domain:ns>
            {% for host in remove_hosts %}
            <domain:hostObj>{{ host }}</domain:hostObj>
            {% endfor %}
          </domain:ns>
          {% endif %}
        </domain:rem>
        {% endif %}
        {% if change %}
        <domain:chg>
          {% if registrant %}
          <domain:registrant>{{ registrant }}</domain:registrant>
          {% endif %}
          {% if password %}
          <domain:authInfo>
            <domain:pw>{{ password }}</domain:pw>
          </domain:authInfo>
          {% endif %}
        </domain:chg>
        {% endif %}
      </domain:update>
    </update>
    <clTRID>{{ client_transaction_id }}</clTRID>
  </command>
</epp>"""

HOST_CHECK_XML = """<?xml version="1.0" encoding="UTF-8" standalone="no"?>
<epp xmlns="urn:ietf:params:xml:ns:epp-1.0">
  <command>
    <check>
      <host:check xmlns:host="urn:ietf:params:xml:ns:host-1.0">
        {% for host_name in host_names %}
        <host:name>{{ host_name }}</host:name>
        {% endfor %}
      </host:check>
    </check>
    <clTRID>{{ client_transaction_id }}</clTRID>
  </command>
</epp>"""

HOST_INFO_XML = """<?xml version="1.0" encoding="UTF-8" standalone="no"?>
<epp xmlns="urn:ietf:params:xml:ns:epp-1.0">
 <command>
   <info>
     <host:info xmlns:host="urn:ietf:params:xml:ns:host-1.0">
       <host:name>{{ host_name }}</host:name>
     </host:info>
   </info>
   <clTRID>{{ client_transaction_id }}</clTRID>
 </command>
</epp>"""

HOST_CREAT_XML = """<?xml version="1.0" encoding="UTF-8" standalone="no"?>
<epp xmlns="urn:ietf:params:xml:ns:epp-1.0">
  <command>
    <create>
      <host:create
       xmlns:host="urn:ietf:params:xml:ns:host-1.0">
        <host:name>{{ host_name }}</host:name>
        {% for addr in address %}
        <host:addr ip="{{ addr.ip }}">{{ addr.address }}</host:addr>
        {% endfor %}
      </host:create>
    </create>
    <clTRID>{{ client_transaction_id }}</clTRID>
  </command>
</epp>"""

HOST_DELETE_XML = """<?xml version="1.0" encoding="UTF-8" standalone="no"?>
<epp xmlns="urn:ietf:params:xml:ns:epp-1.0">
 <command>
   <delete>
     <host:delete xmlns:host="urn:ietf:params:xml:ns:host-1.0">
       <host:name>{{ host_name }}</host:name>
     </host:delete>
   </delete>
   <clTRID>{{ client_transaction_id }}</clTRID>
 </command>
</epp>"""

HOST_UPDATE_XML = """<?xml version="1.0" encoding="UTF-8" standalone="no"?>
<epp xmlns="urn:ietf:params:xml:ns:epp-1.0">
 <command>
   <update>
     <host:update xmlns:host="urn:ietf:params:xml:ns:host-1.0">
       <host:name>{{ host_name }}</host:name>
       {% if add %}
       <host:add>
        {% for addr in add_ip_address %}
        <host:addr ip="{{ addr.ip }}">{{ addr.address }}</host:addr>
        {% endfor %}
        {% for s in add_status %}
         <host:status s="{{ s }}"/>
        {% endfor %}
       </host:add>
       {% endif %}
       {% if remove %}
       <host:rem>
        {% for addr in remove_ip_address %}
        <host:addr ip="{{ addr.ip }}">{{ addr.address }}</host:addr>
        {% endfor %}
        {% for s in remove_status %}
         <host:status s="{{ s }}"/>
        {% endfor %}
       </host:rem>
       {% endif %}
       {% if change %}
       <host:chg>
         <host:name>{{ new_host_name }}</host:name>
       </host:chg>
       {% endif %}
     </host:update>
   </update>
   <clTRID>{{ client_transaction_id }}</clTRID>
 </command>
</epp>"""

POLL_REQUEST_XML = """<?xml version="1.0" encoding="UTF-8" standalone="no"?>
    <epp xmlns="urn:ietf:params:xml:ns:epp-1.0">
      <command>
        <poll op="req"/>
        <clTRID>{{ client_transaction_id }}</clTRID>
      </command>
    </epp>"""

POLL_ACK_XML = """<?xml version="1.0" encoding="UTF-8" standalone="no"?>
<epp xmlns="urn:ietf:params:xml:ns:epp-1.0">
  <command>
    <poll msgID="{{ message_id }}" op="ack"/>
    <clTRID>{{ client_transaction_id }}</clTRID>
  </command>
</epp>"""


"""
Domain cli module
"""
from datetime import datetime

import click

from pyepp.cli import utils
from pyepp.domain import Domain, DomainData


@click.group(name='domain')
@click.pass_context
def domain_group(ctx):
    """To work with Domain name objects in the registry."""
    ctx.obj.registry_object = Domain(ctx.obj.epp)


@click.command(name='info')
@click.argument('domain-name')
@click.option('--client-transaction-id')
@click.pass_context
def domain_info(ctx, domain_name, client_transaction_id) -> None:
    """Returns domain name details."""
    result = ctx.obj.info(domain_name, client_transaction_id)
    utils.echo(result)


@click.command(name='check')
@click.argument('domain-names', nargs=-1)
@click.option('--client-transaction-id')
@click.pass_context
def domain_check(ctx, domain_names: tuple, client_transaction_id) -> None:
    """Checks if domain name(s) exist in the registry."""
    result = ctx.obj.check(list(domain_names), client_transaction_id)
    utils.echo(result)


@click.command(name='delete')
@click.argument('domain-name')
@click.option('--client-transaction-id')
@click.pass_context
def domain_delete(ctx, domain_name, client_transaction_id) -> None:
    """Deletes a given domain from registry."""
    result = ctx.obj.delete(domain_name, client_transaction_id)
    utils.echo(result)


@click.command(name='renew')
@click.argument('domain-name')
@click.argument('current-expiry-date')
@click.option('--period')
@click.option('--client-transaction-id')
@click.pass_context
def domain_renew(ctx, domain_name, current_expiry_date, period, client_transaction_id) -> None:
    """Extend the validity period of a domain name.

    \b
    DOMAIN_NAME: Domain name
    CURRENT_EXPIRY_DATE: The current expiry date in 'YYYY-MM-DD' format.
    """
    result = ctx.obj.renew(domain_name,
                           datetime.strptime(current_expiry_date, "%Y-%m-%d"),
                           period,
                           client_transaction_id)
    utils.echo(result)


@click.command(name='transfer')
@click.argument('domain-name')
@click.argument('password')
@click.option('--period', help="The number of years to be added to the registration period of the domain "
                               "object upon completion of the transfer process.")
@click.option('--client-transaction-id')
@click.pass_context
def domain_transfer(ctx, domain_name, password, period, client_transaction_id) -> None:
    """Extend the validity period of a domain name.

    \b
    DOMAIN_NAME: Domain name
    PASSWORD: The authorization password for the domain object.
    """
    result = ctx.obj.transfer(domain_name,
                              password,
                              period,
                              client_transaction_id)
    utils.echo(result)


@click.command(name='create')
@click.argument('domain-name')
@click.option('--registrant', help="Contact id of the registrant.")
@click.option('--admin', help="Contact id of the administrator of the domain.")
@click.option('--tech', help="Contact id of the technical support of the domain.")
@click.option('--billing', help="Contact id of the billing contact of the domain.")
@click.option('--period', default=1, show_default=True,
              help="The initial registration period of the domain in years.")
@click.option('--ns-host', multiple=True, help='Name server. Can be repeated for multiple name servers.')
@click.option('--client-transaction-id')
@click.pass_context
# pylint: disable=too-many-arguments, too-many-locals
def domain_create(ctx, domain_name, registrant, admin, tech, billing, period, ns_host, client_transaction_id):
    """Create a domain name object in registry.

    DOMAIN_NAME: Domain name
    """
    domain_data = DomainData(
        domain_name=domain_name,
        registrant=registrant,
        admin=admin,
        tech=tech,
        billing=billing,
        period=period,
        host=ns_host
    )
    result = ctx.obj.create(domain_data, client_transaction_id)
    utils.echo(result)


@click.command(name='update')
@click.argument('domain-name')
@click.option('--registrant', help="A contact id to replace the current registrant.")
@click.option('--password', help="A new password to replace the current domain password.")
@click.option('--add-admin', multiple=True, help="Contact id to add a new admin contact to the domain object.")
@click.option('--add-tech', multiple=True, help="Contact id to add a new tech contact to the domain object.")
@click.option('--add-billing', multiple=True, help="Contact id to add a new billing contact to the admin domain.")
@click.option('--add-ns-host', multiple=True, help="Name server host to be added to domain object.")
@click.option('--add-status', type=(str, str), multiple=True,
              help="A status to be added to domain abject. Both status value and description must be provided.")
@click.option('--remove-admin', multiple=True, help="Admin contact id to be removed from domain object.")
@click.option('--remove-tech', multiple=True, help="Technical support contact id to be removed from domain object.")
@click.option('--remove-billing', multiple=True, help="Billing contact id to be removed from domain object.")
@click.option('--remove-ns-host', multiple=True, help="Name server host to be removed from domain object.")
@click.option('--remove-status', multiple=True, help="A status to be removed to domain abject.")
@click.option('--client-transaction-id')
@click.pass_context
# pylint: disable=too-many-arguments, too-many-locals
def domain_update(ctx, domain_name, password, registrant, add_admin, add_tech, add_billing, add_ns_host, add_status,
                  remove_admin, remove_tech, remove_billing, remove_ns_host, remove_status, client_transaction_id):
    """Modify the attributes of a domain object in registry.

    DOMAIN_NAME: Domain name
    """
    result = ctx.obj.update(domain_name=domain_name,
                            registrant=registrant,
                            password=password,
                            add_admins=add_admin,
                            remove_admins=remove_admin,
                            add_techs=add_tech,
                            remove_techs=remove_tech,
                            add_billings=add_billing,
                            remove_billings=remove_billing,
                            add_statues=add_status,
                            remove_statues=remove_status,
                            add_hosts=add_ns_host,
                            remove_hosts=remove_ns_host,
                            client_transaction_id=client_transaction_id)
    utils.echo(result)


domain_group.add_command(domain_check)
domain_group.add_command(domain_info)
domain_group.add_command(domain_delete)
domain_group.add_command(domain_update)
domain_group.add_command(domain_create)
domain_group.add_command(domain_renew)
domain_group.add_command(domain_transfer)



"""
Domain Mapping Module.
"""
from dataclasses import dataclass, asdict
from enum import Enum
from typing import Optional
from datetime import date

from bs4 import BeautifulSoup

from pyepp.epp import EppResultData
from pyepp.base_command import BaseCommand
from pyepp.command_templates import DOMAIN_CHECK_XML, DOMAIN_INFO_XML, DOMAIN_CREATE_XML, DOMAIN_DELETE_XML, \
    DOMAIN_RENEW_XML, TRANSFER_REQUEST_XML, DOMAIN_UPDATE_XML
from pyepp.epp import EppResultCode


class DNSSECAlgorithm(Enum):
    """DNSSEC algorithms enumeration."""
    DSA_SHA_1 = 3
    RSA_SHA_1 = 5
    DSA_NSEC3_SHA1 = 6
    RSASHA1_NSEC3_SHA1 = 7
    RSA_SHA_256 = 8
    RSA_SHA_512 = 10
    GOST_R_34_10_2001 = 12
    ECDSA_CURVE_P_256_WITH_SHA_256 = 13
    ECDSA_CURVE_P_384_WITH_SHA_384 = 14
    ED25519 = 15
    ED448 = 16
    PRIVATE_ALGORITHM = 253
    PRIVATE_ALGORITHM_OID = 255


class DigestTypeEnum(Enum):
    """Digest types enumeration."""
    SHA_1 = 1
    SHA_256 = 2
    GOST_R_34_11_94 = 3
    SHA_384 = 4


class DNSKeyFlagEnum(Enum):
    """DNS Key flags enumeration."""
    FLAG_256 = 256
    FLAG_257 = 257


@dataclass
class DSRecordKeyData:
    """DNSSEC Key data enumeration."""
    flag: str
    algorithm: DNSSECAlgorithm
    public_key: str
    protocol: str = 3


@dataclass
class DSRecordData:
    """DNSSEC dataclass."""
    key_tag: int
    algorithm: DNSSECAlgorithm
    digest_type: DigestTypeEnum
    digest: str
    dns_key: Optional[DSRecordKeyData] = None


@dataclass
class DomainData:
    """Domain name dataclass."""
    # pylint: disable=too-many-instance-attributes
    domain_name: str
    period: Optional[int] = None
    registrant: Optional[str] = ''
    admin: Optional[str] = ''
    tech: Optional[str] = ''
    sponsoring_client_id: Optional[str] = ''
    billing: Optional[str] = ''
    status: Optional[list[str]] = None
    host: Optional[list[str]] = None
    create_date: Optional[str] = ''
    creat_client_id: Optional[str] = ''
    update_client_id: Optional[str] = ''
    update_date: Optional[str] = ''
    expiry_date: Optional[str] = ''
    transfer_date: Optional[str] = ''
    password: Optional[str] = ''
    dns_sec: Optional[list[DSRecordData]] = None


class Domain(BaseCommand):
    """
    Epp domain object class is used to create and manage domain names in Registry.
    """

    def _data_to_dict(self, data: DomainData) -> dict:
        """Convert a domain dataclass to a dictionary.

        :param DomainData data: Domain name details

        :return: Domain name details
        :rtype: dict
        """
        data_dict = asdict(data)

        return data_dict

    # pylint: disable=R0801
    def check(self, domain_names: list[str], client_transaction_id: Optional[str] = None) -> EppResultData:
        """A successful Domain Check request determines whether a domain name is available for use and whether a domain
        name registration can be successfully created in the Registry.

        :param list domain_names: List of domain names
        :param client_transaction_id: Client transaction id

        :return: Result object
        :rtype: EppResultData
        """
        result = self.execute(DOMAIN_CHECK_XML, domain_names=domain_names, client_transaction_id=client_transaction_id)

        if int(result.code) != int(EppResultCode.SUCCESS.value):
            return result

        raw_response = BeautifulSoup(result.raw_response, 'xml')
        domains_check_data = raw_response.find_all('cd')

        result_data = {}
        for domain_cd in domains_check_data:
            domain = domain_cd.find('name')
            available = domain.get('avail') in ('true', '1')
            reason = domain_cd.find('reason').text if not available else None
            result_data[domain.text] = {
                'avail': available,
                'reason': reason,
            }

        result.result_data = result_data

        return result

    def info(self, domain_name: str, client_transaction_id: Optional[str] = None) -> EppResultData:
        """A successful Domain Info request retrieves information associated with an existing domain name.

        :param domain_name: Domain name
        :param client_transaction_id: Client transaction id

        :return: Result object
        :rtype: EppResultData
        """
        result = self.execute(DOMAIN_INFO_XML, domain_name=domain_name, client_transaction_id=client_transaction_id)

        if int(result.code) != int(EppResultCode.SUCCESS.value):
            return result

        raw_response = BeautifulSoup(result.raw_response, 'xml')

        result_data = {
            'domain_name': raw_response.find('name').text,
            'sponsoring_client_id': raw_response.find('clID').text,
            'status': [status.get('s') for status in raw_response.find_all('status')],
            'host': [host.text for host in raw_response.find_all('hostObj')] if raw_response.find('ns') else None,
            'registrant': raw_response.find('registrant').text,
            'admin': raw_response.find('contact', {'type': 'admin'}).text,
            'tech': raw_response.find('contact', {'type': 'tech'}).text,
            'billing': [billing.text for billing in raw_response.find_all('contact', {'type': 'billing'})]
            if raw_response.find('contact', {'type': 'billing'}) else None,
            'create_date': raw_response.find('crDate').text if raw_response.find('crDate') else None,
            'creat_client_id': raw_response.find('crID').text if raw_response.find('crID') else None,
            'expiry_date': raw_response.find('exDate').text if raw_response.find('exDate') else None,
            'update_client_id': raw_response.find('upID').text if raw_response.find('upID') else None,
            'update_date': raw_response.find('upDate').text if raw_response.find('upDate') else None,
            'transfer_date': raw_response.find('trDate').text if raw_response.find('trDate') else None,
            'password': raw_response.find('pw').text if raw_response.find('pw') else None,
            'period': None,
            'dns_sec': None,
        }

        dns_sec = []
        all_ds_data = raw_response.find_all('dsData')
        for ds_data in all_ds_data:
            dns_sec.append(DSRecordData(**{
                'key_tag': ds_data.find('keyTag').text,
                'algorithm': ds_data.find('alg').text,
                'digest_type': ds_data.find('digestType').text,
                'digest': ds_data.find('digest').text,
                'dns_key': {
                    'flag': ds_data.find('flags').text,
                    'protocol': ds_data.find('protocol').text,
                    'algorithm': ds_data.find('alg').text,
                    'public_key': ds_data.find('pubKey').text,
                } if ds_data.find('keyData') else None
            }))

        if dns_sec:
            result_data['dns_sec'] = dns_sec

        result.result_data = DomainData(**result_data)

        return result

    def create(self, domain: DomainData, client_transaction_id: Optional[str] = None) -> EppResultData:
        """A successful Domain Create request creates a domain object in the Registry, and also creates relationships
        between the domain name and previously created contacts and hosts.

        :param domain: Domain name details
        :param client_transaction_id: Client transaction id

        :return: Result object
        :rtype: EppResultData
        """
        params = self._data_to_dict(domain)
        params['client_transaction_id'] = client_transaction_id

        result = self.execute(DOMAIN_CREATE_XML, **params)

        return result

    def delete(self, domain_name: str, client_transaction_id: Optional[str] = None) -> EppResultData:
        """This command provides a transform operation that allows a client to delete a domain object

        :param domain_name: Domain Name
        :param client_transaction_id: Client transaction id

        :return: Result object
        :rtype: EppResultData
        """
        result = self.execute(DOMAIN_DELETE_XML, domain_name=domain_name, client_transaction_id=client_transaction_id)

        return result

    def renew(self, domain_name: str, expiry_date: date, period: Optional[int] = 1,
              client_transaction_id: Optional[str] = None) -> EppResultData:
        """A successful Domain Renew request extends the registration period of a domain name.

        :param domain_name: Domain Name
        :param expiry_date: expiry date
        :param period: period
        :param client_transaction_id: Client transaction id

        :return: Result object
        :rtype: EppResultData
        """
        result = self.execute(DOMAIN_RENEW_XML,
                              domain_name=domain_name,
                              expiry_date=expiry_date.strftime("%Y-%m-%d"),
                              period=str(period),
                              client_transaction_id=client_transaction_id)

        return result

    def transfer(self, domain_name: str, password: str, period: Optional[int] = None,
                 client_transaction_id: Optional[str] = None) -> EppResultData:
        """transfers the sponsorship of a domain name from another Registrar to the Registrar
        submitting the request.

        :param domain_name: Domain Name
        :param password: The authorization password for the domain object
        :param period: period
        :param client_transaction_id: Client transaction id

        :return: Result object
        :rtype: EppResultData
        """
        result = self.execute(TRANSFER_REQUEST_XML,
                              domain_name=domain_name,
                              password=password,
                              period=str(period) if period else None,
                              client_transaction_id=client_transaction_id)

        return result

    # pylint: disable=too-many-arguments,too-many-locals
    def update(self, domain_name: str,
               registrant: Optional[str] = None,
               password: Optional[str] = None,
               add_admins: Optional[list[str]] = None,
               remove_admins: Optional[list[str]] = None,
               add_techs: Optional[list[str]] = None,
               remove_techs: Optional[list[str]] = None,
               add_billings: Optional[list[str]] = None,
               remove_billings: Optional[list[str]] = None,
               add_statues: Optional[list[tuple]] = None,
               remove_statues: Optional[list[str]] = None,
               add_hosts: Optional[list[str]] = None,
               remove_hosts: Optional[list[str]] = None,
               client_transaction_id: Optional[str] = None
               ) -> EppResultData:
        """A successful Domain Update request modifies a domain object in the Registry, and may also add or delete
        relationships between the domain name and previously created hosts and contacts.

        :param domain_name: Domain name to be updated
        :param registrant: A contact id to replace the current registrant
        :param add_admins: A list of contact ids to be added to the admin contacts
        :param remove_admins: A list of contact ids to be removed from the admin contacts
        :param add_techs:  A list of contact ids to be added to the techs contacts
        :param remove_techs: A list of contact ids to be removed from the tech contacts
        :param add_billings: A list of contact ids to add to the billing contacts
        :param remove_billings: A list of contact ids to remove from the billing contacts
        :param add_statues: List of statuses to be added to the domain name. The tuple must contain two
            elements. The first one will be the Status Code and the second element will be Descriptions.
        :param remove_statues: A list of statues to be removed from the domain name.
        :param add_hosts: A list of host names to be added to the domain name.
        :param remove_hosts: A list of host names to be removed from the domain name.
        :param password: A new password to replace the old password.
        :param client_transaction_id: Client transaction id

        :return: Result object
        :rtype: EppResultData
        """
        add = bool(add_admins or add_techs or add_billings or add_statues or add_hosts)
        remove = bool(remove_admins or remove_techs or remove_billings or remove_statues or remove_hosts)
        change = bool(registrant or password)

        result = self.execute(DOMAIN_UPDATE_XML,
                              add=add,
                              remove=remove,
                              change=change,
                              domain_name=domain_name,
                              add_admins=add_admins,
                              remove_admins=remove_admins,
                              add_techs=add_techs,
                              remove_techs=remove_techs,
                              add_billings=add_billings,
                              remove_billings=remove_billings,
                              add_statues=add_statues,
                              remove_statues=remove_statues,
                              add_hosts=add_hosts,
                              remove_hosts=remove_hosts,
                              password=password,
                              registrant=registrant,
                              client_transaction_id=client_transaction_id)

        return result



"""
pyepp module unit tests
"""
import unittest

from unittest.mock import MagicMock, patch

from pyepp.command_templates import HELLO_XML
from pyepp.epp import EppCommunicator, EppCommunicatorException, EppResultCode, EppResultData


class PyEPPTests(unittest.TestCase):

    def setUp(self) -> None:
        self.epp_config = {
            "server": "0.0.0.0",
            "port": "700",
            "client_cert": "path_to_cert",
            "client_key": "path_to_key",
        }
        patch("pyepp.epp.socket.socket").start()
        patch("pyepp.epp.ssl.SSLContext").start()

        self.maxDiff = None

    @patch.object(EppCommunicator, "_read")
    def test_connect(self, mock_read) -> None:
        mock_read.return_value = "greeting"
        epp = EppCommunicator(**self.epp_config)
        greeting = epp.connect()
        self.assertEqual(greeting, "greeting")

    @patch.object(EppCommunicator, "_read")
    def test_connect_exception(self, mock_read) -> None:
        mock_read.side_effect = Exception
        epp = EppCommunicator(**self.epp_config)

        self.assertRaises(EppCommunicatorException, epp.connect)

    @patch.object(EppCommunicator, "_read")
    def test_hello(self, mock_read) -> None:
        mock_read.return_value = "greeting"
        epp = EppCommunicator(**self.epp_config)
        epp.connect()
        hello = epp.hello()
        self.assertEqual(hello, "greeting")

    def test_login_1000(self) -> None:
        epp = EppCommunicator(**self.epp_config)
        expected_result = \
            EppResultData(**{'code': 1000, 'message': "Command completed successfully",
                             'reason': None, 'raw_response': "response", 'result_data': None})
        epp.execute = MagicMock(return_value=expected_result)
        epp.connect = MagicMock(return_value=None)

        result = epp.login("user", "pass")
        self.assertEqual(expected_result, result)

    def test_login_2004(self) -> None:
        epp = EppCommunicator(**self.epp_config)
        expected_result = \
            EppResultData(**{'code': 2004, 'message': "Command completed successfully",
                             'reason': None, 'raw_response': "response", 'result_data': None})
        epp.execute = MagicMock(return_value=expected_result)
        epp.connect = MagicMock(return_value=None)

        self.assertRaises(EppCommunicatorException, epp.login, "user", "pass")

    def test_login_other(self) -> None:
        epp = EppCommunicator(**self.epp_config)
        expected_result = \
            EppResultData(**{'code': 3000, 'message': "Command completed successfully",
                             'reason': None, 'raw_response': "response", 'result_data': None})
        epp.execute = MagicMock(return_value=expected_result)
        epp.connect = MagicMock(return_value=None)

        self.assertRaises(EppCommunicatorException, epp.login, "user", "pass")

    def test_logout(self) -> None:
        epp = EppCommunicator(**self.epp_config)
        expected_result = \
            EppResultData(**{'code': 1500, 'message': "Command completed successfully",
                             'reason': None, 'raw_response': "response", 'result_data': None})
        epp.execute = MagicMock(return_value=expected_result)
        epp.connect = MagicMock(return_value=None)
        epp._socket = MagicMock(close=MagicMock())

        result = epp.logout()
        self.assertEqual(expected_result, result)

    def test_execute_not_connected(self) -> None:
        epp = EppCommunicator(**self.epp_config)
        self.assertRaises(EppCommunicatorException, epp.execute, HELLO_XML)

    def test_execute_no_result(self) -> None:
        epp = EppCommunicator(**self.epp_config)
        epp.connect = MagicMock(return_value=None)
        raw_response = b'<?xml version="1.0" encoding="UTF-8"?>\n<epp xmlns="urn:ietf:params:xml:ns:epp-1.0" xmlns:host="urn:ietf:params:xml:ns:host-1.0" xmlns:contact="urn:ietf:params:xml:ns:contact-1.0" xmlns:domain="urn:ietf:params:xml:ns:domain-1.0" xmlns:rgp="urn:ietf:params:xml:ns:rgp-1.0" xmlns:secDNS="urn:ietf:params:xml:ns:secDNS-1.1">\n    <response>\n            <msg>Command completed successfully</msg>\n        </result>\n        <trID>\n            <svTRID>CIRA-000057351729-0000000001</svTRID>\n        </trID>\n    </response>\n</epp>'
        epp._execute_command = MagicMock(return_value=raw_response)
        self.assertRaises(EppCommunicatorException, epp.execute, HELLO_XML)

    def test_execute_with_error(self) -> None:
        epp = EppCommunicator(**self.epp_config)
        epp.connect = MagicMock(return_value=None)
        epp.greeting = "greeting"
        raw_response = b'''<?xml version="1.0" encoding="UTF-8"?>
<epp xmlns="urn:ietf:params:xml:ns:epp-1.0" xmlns:contact="urn:ietf:params:xml:ns:contact-1.0" xmlns:domain="urn:ietf:params:xml:ns:domain-1.0" xmlns:host="urn:ietf:params:xml:ns:host-1.0" xmlns:rgp="urn:ietf:params:xml:ns:rgp-1.0" xmlns:secDNS="urn:ietf:params:xml:ns:secDNS-1.1"><response><result code="2000"><msg>There is an error</msg><reason>Because!</reason></result><trID><svTRID>CIRA-000057351729-0000000001</svTRID></trID></response></epp>'''
        epp._execute_command = MagicMock(return_value=raw_response)
        expected_result = \
            EppResultData(**{'code': 2000,
                             'message': 'There is an error',
                             'reason': 'Because!',
                             'raw_response': raw_response,
                             'client_transaction_id': None,
                             'server_transaction_id': 'CIRA-000057351729-0000000001',
                             'repository_object_id': None,
                             'result_data': None
                             })

        epp.connect()
        result = epp.execute(HELLO_XML)
        self.assertEqual(result, expected_result)

    def test_execute(self) -> None:
        epp = EppCommunicator(**self.epp_config)
        epp.connect = MagicMock(return_value=None)
        epp.greeting = "greeting"
        raw_response = b'<?xml version="1.0" encoding="UTF-8"?>\n<epp xmlns="urn:ietf:params:xml:ns:epp-1.0" xmlns:host="urn:ietf:params:xml:ns:host-1.0" xmlns:contact="urn:ietf:params:xml:ns:contact-1.0" xmlns:domain="urn:ietf:params:xml:ns:domain-1.0" xmlns:rgp="urn:ietf:params:xml:ns:rgp-1.0" xmlns:secDNS="urn:ietf:params:xml:ns:secDNS-1.1">\n    <response>\n<result code="1000">\n<msg>Command completed successfully</msg>\n</result>\n<trID>\n<svTRID>CIRA-000057351729-0000000001</svTRID>\n</trID>\n</response>\n</epp>'
        epp._execute_command = MagicMock(return_value=raw_response)
        expected_result = \
            EppResultData(**{'code': 1000, 'message': 'Command completed successfully', 'reason': None,
                             'raw_response': raw_response,
                             'client_transaction_id': None,
                             'server_transaction_id': 'CIRA-000057351729-0000000001',
                             'repository_object_id': None,
                             'result_data': None
                             })

        epp.connect()
        result = epp.execute(HELLO_XML)
        self.assertEqual(result, expected_result)

    def test_execute_epp_exception(self) -> None:
        epp = EppCommunicator(**self.epp_config)
        epp.connect = MagicMock(return_value=None)
        epp.greeting = "greeting"
        epp._execute_command = MagicMock(side_effect=EppCommunicatorException)

        self.assertRaises(EppCommunicatorException, epp.execute, HELLO_XML)

    def test_execute_general_exception(self) -> None:
        epp = EppCommunicator(**self.epp_config)
        epp.connect = MagicMock(return_value=None)
        epp.greeting = "greeting"
        epp._execute_command = MagicMock(side_effect=Exception)

        self.assertRaises(EppCommunicatorException, epp.execute, HELLO_XML)

    def test_read(self) -> None:
        epp = EppCommunicator(**self.epp_config)
        expected_result = b'<?xml version="1.0" encoding="UTF-8"?>\n<epp xmlns="urn:ietf:params:xml:ns:epp-1.0" xmlns:fury="urn:ietf:params:xml:ns:fury-1.0" xmlns:droplist="urn:ietf:params:xml:ns:droplist-1.0" xmlns:idn="urn:ietf:params:xml:ns:idn-1.0" xmlns:host="urn:ietf:params:xml:ns:host-1.0" xmlns:contact="urn:ietf:params:xml:ns:contact-1.0" xmlns:domain="urn:ietf:params:xml:ns:domain-1.0" xmlns:secDNS="urn:ietf:params:xml:ns:secDNS-1.1" xmlns:launch="urn:ietf:params:xml:ns:launch-1.0" xmlns:mark="urn:ietf:params:xml:ns:mark-1.0" xmlns:smd="urn:ietf:params:xml:ns:signedMark-1.0" xmlns:ds="http://www.w3.org/2000/09/xmldsig#" xmlns:rgp="urn:ietf:params:xml:ns:rgp-1.0" xmlns:fee="urn:ietf:params:xml:ns:fee-0.11">\n    <greeting>\n        <svID>EPP Server Version: 8.0.4 (de4e6cbdab9536033563b335e4fa4555cf24c931)</svID>\n        <svDate>2023-03-30T20:56:22.740Z</svDate>\n        <svcMenu>\n            <version>1.0</version>\n            <lang>fr</lang>\n            <lang>en</lang>\n            <objURI>urn:ietf:params:xml:ns:epp-1.0</objURI>\n            <objURI>urn:ietf:params:xml:ns:domain-1.0</objURI>\n            <objURI>urn:ietf:params:xml:ns:host-1.0</objURI>\n            <objURI>urn:ietf:params:xml:ns:contact-1.0</objURI>\n            <svcExtension>\n                <extURI>urn:ietf:params:xml:ns:rgp-1.0</extURI>\n                <extURI>urn:ietf:params:xml:ns:fury-1.0</extURI>\n                <extURI>urn:ietf:params:xml:ns:fury-2.0</extURI>\n                <extURI>urn:ietf:params:xml:ns:droplist-1.0</extURI>\n                <extURI>urn:ietf:params:xml:ns:fury-rgp-1.0</extURI>\n                <extURI>urn:ietf:params:xml:ns:idn-1.0</extURI>\n                <extURI>urn:ietf:params:xml:ns:secDNS-1.1</extURI>\n                <extURI>urn:ietf:params:xml:ns:launch-1.0</extURI>\n                <extURI>urn:ietf:params:xml:ns:mark-1.0</extURI>\n                <extURI>urn:ietf:params:xml:ns:signedMark-1.0</extURI>\n                <extURI>urn:ietf:params:xml:ns:fee-0.11</extURI>\n                <extURI>urn:ietf:params:xml:ns:fee-0.9</extURI>\n                <extURI>http://www.w3.org/2000/09/xmldsig#</extURI>\n            </svcExtension>\n        </svcMenu>\n        <dcp>\n            <access>\n                <none/>\n            </access>\n            <statement>\n                <purpose>\n                    <admin/>\n                </purpose>\n                <recipient>\n                    <ours/>\n                </recipient>\n                <retention>\n                    <legal/>\n                </retention>\n            </statement>\n        </dcp>\n    </greeting>\n</epp>'
        epp._ssl_socket = MagicMock(
            read=MagicMock(return_value=b'\x00\x00\n\t'),
            recv=MagicMock(return_value=expected_result)
        )

        result = epp._read()
        self.assertEqual(result, expected_result)

    def test_write(self) -> None:
        epp = EppCommunicator(**self.epp_config)
        expected_result = 103
        epp._ssl_socket = MagicMock(
            send=MagicMock(return_value=expected_result),
        )

        result = epp._write(HELLO_XML)

        self.assertEqual(result, expected_result)


if __name__ == "__main__":
    unittest.main()



"""
EPP Communicator Module
"""
import ssl
import socket
import struct
import logging
import sys
from dataclasses import dataclass, asdict
from enum import Enum
from typing import Optional, Any

from bs4 import BeautifulSoup

from pyepp.command_templates import LOGOUT_XML, LOGIN_XML, HELLO_XML, template_engine

LENGTH_FIELD_SIZE = 4
CRLF_SIZE = 2


class EppCommunicatorException(Exception):
    """
    EPP communicator exception.
    """


class EppResultCode(Enum):
    """
    EPP result codes enumeration.
    """
    SUCCESS = 1000
    SUCCESS_ACTION_PENDING = 1001
    SUCCESS_NO_MESSAGE = 1300
    SUCCESS_ACK_TO_DEQUEUE = 1301
    SUCCESS_END_SESSION = 1500
    PARAMETER_RANGE_ERROR = 2004


def get_format_32() -> str:
    """
    Get the size of C integers. We need 32 bits unsigned.

    From http://www.bortzmeyer.org/4934.html
    """
    format_32 = ">I"
    if struct.calcsize(format_32) < LENGTH_FIELD_SIZE:
        format_32 = ">L"
        if struct.calcsize(format_32) != LENGTH_FIELD_SIZE:
            logging.error("Integer size does not match the length size!")
            raise EppCommunicatorException("Integer size does not match the length size!")
    elif struct.calcsize(format_32) > LENGTH_FIELD_SIZE:
        format_32 = ">H"
        if struct.calcsize(format_32) != LENGTH_FIELD_SIZE:
            logging.error("Integer size does not match the length size!")
            raise EppCommunicatorException("Integer size does not match the length size!")

    return format_32


# pylint: disable=too-many-instance-attributes
@dataclass
class EppResultData:
    """Epp result data structure.
    """
    code: int
    message: str
    raw_response: str
    result_data: Any
    reason: Optional[str] = None
    client_transaction_id:  Optional[str] = None
    server_transaction_id:  Optional[str] = None
    repository_object_id:  Optional[str] = None

    def __setitem__(self, key, value):
        self.__dict__[key] = value

    def __getitem__(self, key):
        return self.__dict__[key]

    def __len__(self):
        return len(self.__dict__)

    def to_dict(self) -> dict:
        """Convert an EppResultData object to a dictionary.
        """
        return asdict(self)


class EppCommunicator:
    """
    An EPP client for connecting to EPP server.
    """

    # pylint: disable=too-many-instance-attributes,too-many-arguments
    def __init__(self, server: str, port: str, client_cert: str,
                 client_key: str, dry_run: Optional[bool] = False) -> None:
        """
        :param server: EPP server to connect to.
        :param port: EPP port to connect to.
        :param client_cert: Path client certificate
        :param client_key: Path to client key
        :param dry_run: dry run the request
        """
        self._server = server
        self._port = port
        self._user = None
        self._client_cert = client_cert
        self._client_key = client_key

        self._dry_run = dry_run

        self._format_32 = get_format_32()

        self._context = None
        self._socket = None
        self._ssl_socket = None
        self.greeting = None

    @property
    def user(self):
        """User property
        """
        return self._user

    def _unpack_data(self, data: int) -> str:
        """
        Unpack data.

        :param bytes data: data

        :return: unpacked data
        :rtype: str
        """
        return struct.unpack(self._format_32, data)[0]

    def _pack_data(self, data: int) -> bytes:
        """
        Pack the data.

        :param data: data

        :return: bytes
        """
        return struct.pack(self._format_32, data)

    def _read(self) -> bytes:
        """
        Read the response from the socket.

        :return: Response
        :rtype: bytes
        """
        length = self._ssl_socket.read(LENGTH_FIELD_SIZE)
        buffer = bytes()

        if not length:
            return None

        total_bytes = self._unpack_data(length) - LENGTH_FIELD_SIZE
        while len(buffer) < total_bytes:
            total_bytes = total_bytes - len(buffer)
            buffer += self._ssl_socket.recv(total_bytes)
            logging.info('Received %s/%s bytes', len(buffer), total_bytes)
        return buffer

    def _write(self, xml: str) -> int:
        """
        Write the request into the socket.

        :param str xml: XML Command

        :return: Number of send bytes
        :rtype: int
        """
        # +4 for the length field itself (section 4 mandates that)
        # +2 for the CRLF at the end
        length = self._pack_data(len(xml) + LENGTH_FIELD_SIZE + CRLF_SIZE)

        self._ssl_socket.send(length)
        xml += "\r\n"
        return self._ssl_socket.send(xml.encode("utf-8"))

    def _execute_command(self, cmd: str) -> bytes:
        """
        Execute the command. Sending the request to the server and receive the response.

        :param str cmd: XML command

        :return: Response
        :rtype: bytes
        """

        # Print the xml command and exit the app
        if self._dry_run:
            print(cmd)
            sys.exit()

        logging.debug("Sending xml to server :\n%s", cmd)

        self._write(cmd)

        response = self._read()
        if response is None:
            raise EppCommunicatorException("Cannot connect to server. Please re-login!")

        logging.debug("Received xml response from server :\n%s", response)

        return response

    def connect(self) -> bytes:
        """
        Initial connect to the server.

        :return: Greeting message
        :rtype: bytes

        :raises EppCommunicatorException: When there is any errors
        """
        try:
            self._context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
            self._context.load_default_certs()
            self._context.load_cert_chain(certfile=self._client_cert, keyfile=self._client_key)

            self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0)
            self._socket.settimeout(10)

            self._ssl_socket = self._context.wrap_socket(self._socket, server_hostname=self._server)
            self._ssl_socket.connect((self._server, int(self._port)))
            self.greeting = self._read()
            logging.debug(BeautifulSoup(self.greeting, 'xml'))
            return self.greeting
        except Exception as ex:
            logging.error("Could not setup a sec sure connection. %s", str(ex))
            raise EppCommunicatorException("Could not setup a sec sure connection") from ex

    def execute(self, cmd: str) -> EppResultData:
        """
        Execute the command. Sending the request to the server and receive the response.

        :param str cmd: XML Command

        :return: Result object
        :rtype: EppResultData

        :raises EppCommunicatorException: When there is any errors.
        """
        try:
            if not self.greeting and not self._dry_run:
                raise EppCommunicatorException("The connection to the server has not been established yet!")

            raw_response = self._execute_command(cmd)
            xml_response = BeautifulSoup(raw_response, 'xml')

            response = xml_response.find('response')
            result = xml_response.find('result')
            message = result.find('msg').string

            try:
                code = int(result.get('code'))
            except AttributeError as exc:
                raise EppCommunicatorException("Could not get result code.") from exc

            reason = None
            if code not in (EppResultCode.SUCCESS.value, EppResultCode.SUCCESS_END_SESSION.value):
                reason = result.find('reason').string if result.find('reason') else None

            client_transaction_id = response.find('clTRID').text if response.find('clTRID') else None
            server_transaction_id = response.find('svTRID').text if response.find('svTRID') else None
            repository_object_id = response.find('roid').text if response.find('roid') else None

            logging.debug("Command executed:\n%s", xml_response)

            return EppResultData(
                code=code,
                message=message,
                reason=reason,
                raw_response=raw_response,
                client_transaction_id=client_transaction_id,
                server_transaction_id=server_transaction_id,
                repository_object_id=repository_object_id,
                result_data=None
            )
        except EppCommunicatorException as epp_ex:
            raise epp_ex
        except Exception as ex:
            raise EppCommunicatorException(ex) from ex

    def hello(self) -> bytes:
        """
        Send Hello command the server.

        :return: Greeting response
        :rtype: bytes
        """
        logging.debug("Send Hello command to the server!")
        greeting = self._execute_command(HELLO_XML)
        return greeting

    def login(self, user: str, password: str, extensions: Optional[list[str]] = None) -> EppResultData:
        """
        Login the user to EPP server.

        :param user: username
        :param password: password
        :param extensions: A list of supported extension URIs

        :return: Result object
        :rtype: EppResultData

        :raises EppCommunicatorException: When there are any errors.
        """
        if extensions is None:
            extensions = []

        self._user = user

        command_template = template_engine.from_string(LOGIN_XML)
        command = command_template.render(user=user, password=password, extensions=extensions)

        result = self.execute(command)

        if result.code == EppResultCode.SUCCESS.value:
            logging.info("User %s logged in to %s:%s", self._user, self._server, self._port)
        elif result.code == EppResultCode.PARAMETER_RANGE_ERROR.value:
            raise EppCommunicatorException("Incorrect user name or password. Please try again!")
        else:
            raise EppCommunicatorException(f"Something went wrong! Code: {result.code} - Message: "
                                           f"{result.message} - Reason {result.reason}")

        return result

    def logout(self) -> EppResultData:
        """
        Logout the user from EPP server.

        :return: Result object
        :rtype: EppResultData
        """

        logout = self.execute(LOGOUT_XML)
        self._socket.close()
        logging.info("User %s logged out from %s:%s", self._user, self._server, self._port)

        return logout



"""
Host Name Mapping unit tests
"""
import unittest
from datetime import date
from unittest.mock import MagicMock

from pyepp.host import Host, HostData, IPAddressData
from pyepp.epp import EppCommunicator, EppResultData


class HostTest(unittest.TestCase):
    """
    Host mapping name unit tests
    """

    def setUp(self) -> None:
        self.maxDiff = None

    def test_check_unsuccessful(self) -> None:
        epp_communicator = MagicMock(EppCommunicator)
        host = Host(epp_communicator)
        expected_result = \
            EppResultData(**{'code': 2000, 'message': "Command completed unsuccessfully",
                             'reason': None, 'raw_response': "response", 'result_data': None})
        host.execute = MagicMock(return_value=expected_result)

        result = host.check(['test.host.nz'])

        self.assertEqual(result, expected_result)

    def test_check(self) -> None:
        epp_communicator = MagicMock(EppCommunicator)
        host = Host(epp_communicator)
        expected_result = EppResultData(**{'client_transaction_id': 'cc7d1dba-2c89-4934-9cd2-688f14aafc8c',
                           'code': 1000,
                           'message': 'Command completed successfully',
                           'raw_response': '<response>\n'
                                           '<result code="1000">\n'
                                           '<msg>Command completed successfully</msg>\n'
                                           '</result>\n'
                                           '<resData>\n'
                                           '<host:chkData>\n'
                                           '<host:cd>\n'
                                           '<host:name avail="true">test1.ehsan.nz</host:name>\n'
                                           '</host:cd>\n'
                                           '<host:cd>\n'
                                           '<host:name '
                                           'avail="false">0qx.test-3gudmj2badfgwhaubrm8douuldtkz4.co.nz</host:name>\n'
                                           '<host:reason>Selected host is not available</host:reason>\n'
                                           '</host:cd>\n'
                                           '</host:chkData>\n'
                                           '</resData>\n'
                                           '<trID>\n'
                                           '<clTRID>cc7d1dba-2c89-4934-9cd2-688f14aafc8c</clTRID>\n'
                                           '<svTRID>CIRA-000071511943-0000000003</svTRID>\n'
                                           '</trID>\n'
                                           '</response>',
                           'reason': None,
                           'repository_object_id': None,
                           'result_data': {'0qx.test-3gudmj2badfgwhaubrm8douuldtkz4.co.nz': {'avail': False,
                                                                                             'reason': 'Selected '
                                                                                                       'host '
                                                                                                       'is '
                                                                                                       'not '
                                                                                                       'available'},
                                           'test1.host.nz': {'avail': True, 'reason': None}},
                           'server_transaction_id': 'CIRA-000071511943-0000000003'})
        host.execute = MagicMock(return_value=expected_result)

        result = host.check(['0qx.test-3gudmj2badfgwhaubrm8douuldtkz4.co.nz', 'itest1.host.nz'])

        self.assertEqual(result, expected_result)

    def test_info_unsuccessful(self) -> None:
        epp_communicator = MagicMock(EppCommunicator)
        host = Host(epp_communicator)
        expected_result = \
            EppResultData(**{'code': 2000, 'message': "Command completed unsuccessfully",
                             'reason': None, 'raw_response': "response", 'result_data': None})
        host.execute = MagicMock(return_value=expected_result)

        result = host.info('host.internet.nz')

        self.assertEqual(result, expected_result)

    def test_info(self) -> None:
        epp_communicator = MagicMock(EppCommunicator)
        host = Host(epp_communicator)
        execute_result = (
            EppResultData(**{'client_transaction_id': 'fecb12df-bc07-4acf-bbc8-e2c5326f63e8',
                             'code': 1000,
                             'message': 'Command completed successfully',
                             'raw_response': '<response>\n'
                                             '<result code="1000">\n'
                                             '<msg>Command completed successfully</msg>\n'
                                             '</result>\n'
                                             '<resData>\n'
                                             '<host:infData>\n'
                                             '<host:name>0qx.test-3gudmj2badfgwhaubrm8douuldtkz4.co.nz</host:name>\n'
                                             '<host:roid>12313305-INZ</host:roid>\n'
                                             '<host:status s="linked"/>\n'
                                             '<host:addr ip="v4">192.1.2.3</host:addr>\n'
                                             '<host:addr '
                                             'ip="v6">2620:0:1009:3:7426:52dc:b8c1:51b2</host:addr>\n'
                                             '<host:clID>163</host:clID>\n'
                                             '<host:crID>163</host:crID>\n'
                                             '<host:crDate>2023-05-27T08:06:36.407Z</host:crDate>\n'
                                             '<host:upID>CIRA_RAR_1</host:upID>\n'
                                             '<host:upDate>2023-05-27T08:09:49.271Z</host:upDate>\n'
                                             '</host:infData>\n'
                                             '</resData>\n'
                                             '<trID>\n'
                                             '<clTRID>fecb12df-bc07-4acf-bbc8-e2c5326f63e8</clTRID>\n'
                                             '<svTRID>CIRA-000073271502-0000000003</svTRID>\n'
                                             '</trID>\n'
                                             '</response>',
                             'reason': None,
                             'repository_object_id': '12313305-INZ',
                             'server_transaction_id': 'CIRA-000073271502-0000000003',
                             'result_data': None}))

        expected_result = (
            EppResultData(**{'client_transaction_id': 'fecb12df-bc07-4acf-bbc8-e2c5326f63e8',
                             'code': 1000,
                             'message': 'Command completed successfully',
                             'raw_response': '<response>\n'
                                             '<result code="1000">\n'
                                             '<msg>Command completed successfully</msg>\n'
                                             '</result>\n'
                                             '<resData>\n'
                                             '<host:infData>\n'
                                             '<host:name>0qx.test-3gudmj2badfgwhaubrm8douuldtkz4.co.nz</host:name>\n'
                                             '<host:roid>12313305-INZ</host:roid>\n'
                                             '<host:status s="linked"/>\n'
                                             '<host:addr ip="v4">192.1.2.3</host:addr>\n'
                                             '<host:addr '
                                             'ip="v6">2620:0:1009:3:7426:52dc:b8c1:51b2</host:addr>\n'
                                             '<host:clID>163</host:clID>\n'
                                             '<host:crID>163</host:crID>\n'
                                             '<host:crDate>2023-05-27T08:06:36.407Z</host:crDate>\n'
                                             '<host:upID>CIRA_RAR_1</host:upID>\n'
                                             '<host:upDate>2023-05-27T08:09:49.271Z</host:upDate>\n'
                                             '</host:infData>\n'
                                             '</resData>\n'
                                             '<trID>\n'
                                             '<clTRID>fecb12df-bc07-4acf-bbc8-e2c5326f63e8</clTRID>\n'
                                             '<svTRID>CIRA-000073271502-0000000003</svTRID>\n'
                                             '</trID>\n'
                                             '</response>',
                             'reason': None,
                             'repository_object_id': '12313305-INZ',
                             'result_data': HostData(host_name='0qx.test-3gudmj2badfgwhaubrm8douuldtkz4.co.nz',
                                                     status=['linked'],
                                                     address=[IPAddressData(address='192.1.2.3', ip='v4'),
                                                              IPAddressData(address='2620:0:1009:3:7426:52dc:b8c1:51b2',
                                                                            ip='v6')],
                                                     create_date='2023-05-27T08:06:36.407Z',
                                                     creat_client_id='163',
                                                     update_client_id='CIRA_RAR_1',
                                                     update_date='2023-05-27T08:09:49.271Z'),
                             'server_transaction_id': 'CIRA-000073271502-0000000003'}))

        host.execute = MagicMock(return_value=execute_result)

        result = host.info('0qx.test-3gudmj2badfgwhaubrm8douuldtkz4.co.nz')

        self.assertEqual(result, expected_result)

    def test_create(self) -> None:
        expected_result = (
            EppResultData(**{'client_transaction_id': 'b4450b14-4115-4185-8615-b77f7ad85dab',
                             'code': 1000,
                             'message': 'Command completed successfully',
                             'raw_response': '<response>\n'
                                             '<result code="1000">\n'
                                             '<msg>Command completed successfully</msg>\n'
                                             '</result>\n'
                                             '<resData>\n'
                                             '<host:creData>\n'
                                             '<host:name>host.inz.nz</host:name>\n'
                                             '<host:crDate>2023-06-22T00:08:32.739Z</host:crDate>\n'
                                             '</host:creData>\n'
                                             '</resData>\n'
                                             '<trID>\n'
                                             '<clTRID>b4450b14-4115-4185-8615-b77f7ad85dab</clTRID>\n'
                                             '<svTRID>CIRA-000074059948-0000000003</svTRID>\n'
                                             '</trID>\n'
                                             '</response>',
                             'reason': None,
                             'repository_object_id': None,
                             'server_transaction_id': 'CIRA-000074059948-0000000003',
                             'result_data': None}))

        create_params = HostData(host_name='host.inz.nz', address=[IPAddressData(address='192.168.1.100', ip='v4'),
                                                                   IPAddressData(address='192.168.1.200', ip='v4')])
        epp_communicator = MagicMock(EppCommunicator)
        host = Host(epp_communicator)
        host.execute = MagicMock(return_value=expected_result)

        result = host.create(create_params)

        self.assertEqual(result, expected_result)

    def test_delete(self) -> None:
        expected_result = (
            EppResultData(**{'client_transaction_id': 'b9c0c77d-b2d9-47c7-9edb-b4d4988c2e7d',
                             'code': 1000,
                             'message': 'Command completed successfully',
                             'raw_response': '<response>\n'
                                             '<result code="1000">\n'
                                             '<msg>Command completed successfully</msg>\n'
                                             '</result>\n'
                                             '<trID>\n'
                                             '<clTRID>b9c0c77d-b2d9-47c7-9edb-b4d4988c2e7d</clTRID>\n'
                                             '<svTRID>CIRA-000064688525-0000000004</svTRID>\n'
                                             '</trID>\n'
                                             '</response>',
                             'reason': None,
                             'repository_object_id': None,
                             'server_transaction_id': 'CIRA-000064688525-0000000004',
                             'result_data': None}))

        epp_communicator = MagicMock(EppCommunicator)
        host = Host(epp_communicator)
        host.execute = MagicMock(return_value=expected_result)

        result = host.delete('host.internet.nz')

        self.assertEqual(result, expected_result)

    def test_update(self) -> None:
        expected_result = (
            EppResultData(**{'client_transaction_id': '826aa351-2207-4582-846d-7839630f9c4f',
                             'code': 1000,
                             'message': 'Command completed successfully',
                             'raw_response': '<response>\n'
                                             '<result code="1000">\n'
                                             '<msg>Command completed successfully</msg>\n'
                                             '</result>\n'
                                             '<trID>\n'
                                             '<clTRID>826aa351-2207-4582-846d-7839630f9c4f</clTRID>\n'
                                             '<svTRID>CIRA-000077350909-0000000003</svTRID>\n'
                                             '</trID>\n'
                                             '</response>',
                             'reason': None,
                             'repository_object_id': None,
                             'server_transaction_id': 'CIRA-000077350909-0000000003',
                             'result_data': None}))

        epp_communicator = MagicMock(EppCommunicator)
        host = Host(epp_communicator)
        host.execute = MagicMock(return_value=expected_result)

        result = host.update(host_name='host.internet.nz',
                             remove_ip_address=[IPAddressData(address='192.168.1.10', ip='v4')],
                             add_ip_address=[IPAddressData(address='192.168.1.20', ip='v4')],
                             new_host_name='new-host.internet.nz')

        self.assertEqual(result, expected_result)



"""
Helper functions
"""
import random
import string

from bs4 import BeautifulSoup


def generate_password(length: int) -> str:
    """Generate a random password including letters and digits.

    :param int length: password length

    :return: password
    :rtype: str
    """
    return ''.join([random.choice(string.ascii_letters + string.digits) for _ in range(length)])  # nosec


def xml_pretty(bxml: bytes) -> str:
    """
    Convert bytes xml to string and prettify it.

    :param bxml: xml content

    :return: xml in string
    """
    xml_str = BeautifulSoup(bxml, 'xml')
    return xml_str.decode(pretty_print=True)



"""
Host Mapping Module. This module is used to manage host objects in Registry.
A host object represents a Domain Name System (DNS) server that resolves domain names into IP addresses.
"""
from dataclasses import dataclass, asdict
from typing import Optional

from bs4 import BeautifulSoup

from pyepp.epp import EppResultData
from pyepp.base_command import BaseCommand
from pyepp.command_templates import HOST_CHECK_XML, HOST_INFO_XML, HOST_CREAT_XML, HOST_DELETE_XML, HOST_UPDATE_XML
from pyepp.epp import EppResultCode


@dataclass
class IPAddressData:
    """IP Address data class.
    """
    address: str
    # pylint: disable=invalid-name
    ip: str


@dataclass
class HostData:
    """Host object data class.
    """
    host_name: str
    address: Optional[list[IPAddressData]] = None
    status: Optional[list[str]] = None
    create_date: Optional[str] = ''
    creat_client_id: Optional[str] = ''
    update_client_id: Optional[str] = ''
    update_date: Optional[str] = ''


class Host(BaseCommand):
    """
    Epp Host object class. This class is used to create and manage host objects in Registry.

    An EPP host object has attributes and associated values that can be viewed and modified by the sponsoring
    Registrar. When you create or update a host, you must follow the rules that ensure that the domain name can still
    be resolved even if there is an outage of any individual host or network. This means that two or more unique name
    servers must be defined for each protocol in use. Although you can use the IPv4 and IPv6 protocols concurrently,
    they are not interchangeable or compatible. As a result, to use both protocols requires that you must provide data
    for two unique name servers that are accessible by IPv4 and two unique name servers that are accessible by IPv6.
    """

    def _data_to_dict(self, data: HostData) -> dict:
        """Convert a host dataclass to a dictionary.

        :param data: Host name details

        :return: Host name details
        :rtype: dict
        """
        data_dict = asdict(data)

        return data_dict

    def check(self, host_names: list[str], client_transaction_id: Optional[str] = None) -> EppResultData:
        """A successful Host Check request determines whether a host is available for use and whether a host can be
        created in the Registry. A single request can check from 1 to 15 host names.

        :param host_names: List of domain names
        :param client_transaction_id: Client transaction id

        :return: Result object
        :rtype: EppResultData
        """
        result = self.execute(HOST_CHECK_XML, host_names=host_names, client_transaction_id=client_transaction_id)

        if int(result.code) != int(EppResultCode.SUCCESS.value):
            return result

        raw_response = BeautifulSoup(result.raw_response, 'xml')
        domains_check_data = raw_response.find_all('cd')

        result_data = {}
        for domain_cd in domains_check_data:
            domain = domain_cd.find('name')
            available = domain.get('avail') in ('true', '1')
            reason = domain_cd.find('reason').text if not available else None
            result_data[domain.text] = {
                'avail': available,
                'reason': reason,
            }

        result['result_data'] = result_data

        return result

    def info(self, host_name: str, client_transaction_id: Optional[str] = None) -> EppResultData:
        """A successful Host Info request retrieves detailed host information.

        :param host_name: Host name
        :param client_transaction_id: Client transaction id

        :return: Result object
        :rtype: EppResultData
        """
        result = self.execute(HOST_INFO_XML, host_name=host_name, client_transaction_id=client_transaction_id)

        if int(result.code) != int(EppResultCode.SUCCESS.value):
            return result

        raw_response = BeautifulSoup(result.raw_response, 'xml')

        result_data = {
            'host_name': raw_response.find('name').text,
            'status': [status.get('s') for status in raw_response.find_all('status')],
            'address': [IPAddressData(address=addr.text, ip=addr.get('ip')) for addr in raw_response.find_all('addr')]
            if raw_response.find('addr') else None,
            'create_date': raw_response.find('crDate').text if raw_response.find('crDate') else None,
            'creat_client_id': raw_response.find('crID').text if raw_response.find('crID') else None,
            'update_client_id': raw_response.find('upID').text if raw_response.find('upID') else None,
            'update_date': raw_response.find('upDate').text if raw_response.find('upDate') else None,
        }

        result['result_data'] = HostData(**result_data)

        return result

    def create(self, host: HostData, client_transaction_id: Optional[str] = None) -> EppResultData:
        """A successful Host Create request creates a host object identified by its name in the Registry.

        :param host: Contact
        :param client_transaction_id: Client transaction id

        :return: Result object
        :rtype: EppResultData
        """
        params = self._data_to_dict(host)
        params['client_transaction_id'] = client_transaction_id

        result = self.execute(HOST_CREAT_XML, **params)

        return result

    def delete(self, host_name: str, client_transaction_id: Optional[str] = None) -> EppResultData:
        """A successful Host Delete request deletes a host object.
        Warning: A host object cannot be deleted while it is associated with a domain object.

        :param host_name: Host Name
        :param client_transaction_id: Client transaction id

        :return: Result object
        :rtype: EppResultData
        """
        result = self.execute(HOST_DELETE_XML, host_name=host_name, client_transaction_id=client_transaction_id)

        return result

    # pylint: disable=too-many-arguments
    def update(self, host_name: str,
               add_ip_address: Optional[list[IPAddressData]] = None,
               remove_ip_address: Optional[list[IPAddressData]] = None,
               add_status: Optional[list[str]] = None,
               remove_status: Optional[list[str]] = None,
               new_host_name: Optional[str] = None,
               client_transaction_id: Optional[str] = None
               ) -> EppResultData:
        """The EPP <update> command provides a transform operation that allows a client to modify the attributes of a
        host object.

        :param host_name: Host Name
        :param add_ip_address: A list of IP addressed to be added to the host
        :param remove_ip_address: A list of IP addressed to be removed from the host
        :param add_status: A list of status to be added to the host
        :param remove_status: A list of status to be removed from host
        :param new_host_name: The host name will be changed to this new host name
        :param client_transaction_id: Client transaction id

        :return: Result object
        :rtype: EppResultData
        """

        add = bool(add_ip_address or add_status)
        remove = bool(remove_ip_address or remove_status)
        change = bool(new_host_name)

        result = self.execute(HOST_UPDATE_XML,
                              host_name=host_name,
                              add=add,
                              remove=remove,
                              change=change,
                              add_ip_address=add_ip_address,
                              remove_ip_address=remove_ip_address,
                              add_status=add_status,
                              remove_status=remove_status,
                              new_host_name=new_host_name,
                              client_transaction_id=client_transaction_id)

        return result



"""
Base Command unit tests
"""
import unittest
from unittest.mock import MagicMock, patch
from uuid import UUID

from pyepp.base_command import BaseCommand
from pyepp.epp import EppCommunicator


class BaseCommandTest(unittest.TestCase):

    @patch('pyepp.base_command.uuid.uuid4')
    def test_prepare_command_client_transaction_id(self, mock_uuid) -> None:
        command = """<?xml version="1.0" encoding="UTF-8"?>
<epp xmlns="urn:ietf:params:xml:ns:epp-1.0"><command><clTRID>{{ client_transaction_id }}</clTRID></command></epp>"""
        expected_result = """<?xml version="1.0" encoding="UTF-8"?>
<epp xmlns="urn:ietf:params:xml:ns:epp-1.0"><command><clTRID>6d053972-b813-4659-8029-924546e94489</clTRID></command></epp>"""
        mock_uuid.return_value = UUID('6d053972-b813-4659-8029-924546e94489')
        epp_communicator = MagicMock(EppCommunicator)

        base_command = BaseCommand(epp_communicator)
        resul = base_command._prepare_command(command)

        self.assertEqual(expected_result, resul)

    @patch('pyepp.base_command.helper.generate_password')
    def test_prepare_command_password(self, mock_generate_password) -> None:
        command = """<?xml version="1.0" encoding="UTF-8"?>
<epp xmlns="urn:ietf:params:xml:ns:epp-1.0"><command><contact:authInfo><contact:pw>{{ password }}</contact:pw></contact:authInfo></command></epp>"""
        expected_result = """<?xml version="1.0" encoding="UTF-8"?>
<epp xmlns="urn:ietf:params:xml:ns:epp-1.0"><command><contact:authInfo><contact:pw>wle2Dj6zIJCHTNrX</contact:pw></contact:authInfo></command></epp>"""
        mock_generate_password.return_value = 'wle2Dj6zIJCHTNrX'
        epp_communicator = MagicMock(EppCommunicator)

        base_command = BaseCommand(epp_communicator)
        resul = base_command._prepare_command(command)

        self.assertEqual(expected_result, resul)

    def test_escape_list(self) -> None:
        test_list = [
            ['hi&', '<world>'],
            {'key': '<hello&world>'},
            '<hello&world>'
        ]
        expected_result = [
            ['hi&amp;', '&lt;world&gt;'],
            {'key': '&lt;hello&amp;world&gt;'},
            '&lt;hello&amp;world&gt;'
        ]

        epp_communicator = MagicMock(EppCommunicator)
        base_command = BaseCommand(epp_communicator)

        result = base_command._BaseCommand__escape_list(test_list)

        self.assertEqual(expected_result, result)

    def test_escape_dict(self) -> None:
        test_list = {
            'key1': ['hi&', '<world>'],
            'key2': {'key': '<hello&world>'},
            'key3': '<hello&world>'
        }
        expected_result = {
            'key1': ['hi&amp;', '&lt;world&gt;'],
            'key2': {'key': '&lt;hello&amp;world&gt;'},
            'key3': '&lt;hello&amp;world&gt;'
        }

        epp_communicator = MagicMock(EppCommunicator)
        base_command = BaseCommand(epp_communicator)

        result = base_command._BaseCommand__escape_dict(test_list)

        self.assertEqual(expected_result, result)

    def test_execute(self) -> None:
        command = """<?xml version="1.0" encoding="UTF-8" standalone="no"?>
<epp xmlns="urn:ietf:params:xml:ns:epp-1.0">
  <command>
    <create>
      <contact:create xmlns:contact="urn:ietf:params:xml:ns:contact-1.0">
        <contact:id>{{ id }}</contact:id>
          <contact:postalInfo type="loc">
            <contact:name>{{ name }}</contact:name>
            {% if organization %} <contact:org>{{ organization }}</contact:org> {% endif %}
            <contact:addr>
              <contact:street>{{ street_1 }}</contact:street>
              {% if street_2 %} <contact:street>{{ street_2 }}</contact:street> {% endif %}
              {% if street_3 %} <contact:street>{{ street_3 }}</contact:street> {% endif %}
              <contact:city>{{ city }}</contact:city>
              {% if province %} <contact:sp>{{ province }}</contact:sp> {% endif %}
              {% if postal_code %} <contact:pc>{{ postal_code }}</contact:pc> {% endif %}
              <contact:cc>{{ country_code }}</contact:cc>
           </contact:addr>
         </contact:postalInfo>
         {% if phone %} <contact:voice>{{ phone }}</contact:voice> {% endif %}
         {% if fax %} <contact:fax>{{ fax }}</contact:fax> {% endif %}
         <contact:email>{{ email }}</contact:email>
         <contact:authInfo>
           <contact:pw> {{ password }}</contact:pw>
         </contact:authInfo>
         {% if privacy %}
         <contact:disclose flag="0">
           {% for element in privacy %}
           <contact:{{ element }}/>
           {% endfor %}
         </contact:disclose>
         {% endif %}
       </contact:create>
    </create>
    <clTRID>{{ client_transaction_id }}</clTRID>
  </command>
</epp>"""

        expected_result = {
                'code': '1000',
                'message': 'Command completed successfully',
                'reason': None,
                'raw_response': 'raw_response',
                'client_transaction_id': 'client_transaction_id',
                'server_transaction_id': 'server_transaction_id',
                'repository_object_id': 'repository_object_id',
            }

        params = {
            'id': 'contact-1-id',
            'name': 'contact-1',
            'street_1': '18 Willis Street',
            'street_2': 'Wellington CBD',
            'postal_code': '6011',
            'email': 'mail@mail.net',
        }

        epp_communicator = MagicMock(EppCommunicator)
        epp_communicator.execute.return_value = expected_result
        base_command = BaseCommand(epp_communicator)

        result = base_command.execute(command, **params)

        self.assertEqual(expected_result, result)


"""
PyEpp command line interface module.
"""
# pylint: skip-file
import functools
import pprint

import click

from pyepp.helper import xml_pretty
from pyepp.epp import EppResultData
from pyepp import EppCommunicator


def login_logout(func):
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        if not self.dry_run:
            self.connect()
            self.login()

        result = func(self, *args, **kwargs)

        if not self.dry_run:
            self.logout()

        return result

    return wrapper


class PyEppCli:
    def __init__(self, server, port, client_cert, client_key, user, password, output_format, no_pretty, dry_run=False):
        self.epp = EppCommunicator(server, port, client_cert, client_key, dry_run)

        self.user = user
        self.password = password

        self.output_format = output_format
        self.no_pretty = no_pretty
        self.dry_run = dry_run

        self.registry_object = None

    def connect(self):
        self.epp.connect()

    def login(self):
        self.epp.login(self.user, self.password)

    def logout(self):
        self.epp.logout()

    def format_output(self, result: EppResultData):
        output = result
        if self.output_format == 'XML':
            output = result.raw_response
            if not self.no_pretty:
                output = xml_pretty(result.raw_response)
        elif self.output_format == 'MIN':
            return result.result_data
        else:
            if not self.no_pretty:
                output = pprint.pformat(output)

        return output

    @login_logout
    def execute(self, *args, **kwargs):
        result = self.epp.execute(*args, **kwargs)
        return self.format_output(result)

    @login_logout
    def check(self, *args, **kwargs):
        result = self.registry_object.check(*args, **kwargs)
        return self.format_output(result)

    @login_logout
    def info(self, *args, **kwargs):
        result = self.registry_object.info(*args, **kwargs)
        return self.format_output(result)

    @login_logout
    def delete(self, *args, **kwargs):
        result = self.registry_object.delete(*args, **kwargs)
        return self.format_output(result)

    @login_logout
    def update(self, *args, **kwargs):
        result = self.registry_object.update(*args, **kwargs)
        return self.format_output(result)

    @login_logout
    def create(self, *args, **kwargs):
        result = self.registry_object.create(*args, **kwargs)
        return self.format_output(result)

    @login_logout
    def renew(self, *args, **kwargs):
        result = self.registry_object.renew(*args, **kwargs)
        return self.format_output(result)

    @login_logout
    def transfer(self, *args, **kwargs):
        result = self.registry_object.transfer(*args, **kwargs)
        return self.format_output(result)

    @login_logout
    def request(self, *args, **kwargs):
        result = self.registry_object.request(*args, **kwargs)
        return self.format_output(result)

    @login_logout
    def acknowledge(self, *args, **kwargs):
        result = self.registry_object.acknowledge(*args, **kwargs)
        return self.format_output(result)


pass_pyepp_cli = click.make_pass_decorator(PyEppCli)
